@@ -476,9 +476,12 @@ class ChannelState(utils.python.Object):
         return ret
 
 
-Batch = collections.namedtuple('Batch', 'type arguments messages')
+Batch = collections.namedtuple('Batch', 'name type arguments messages parent_batch')
 """Represents a batch of messages, see
-<https://ircv3.net/specs/extensions/batch-3.2>"""
+<https://ircv3.net/specs/extensions/batch-3.2>
+
+Only access attributes by their name and do not create Batch objects
+in plugins; so we can extend the structure without breaking plugins."""
 
 
 class IrcStateFsm(object):
@@ -752,10 +755,11 @@ class IrcState(IrcCommandDispatcher, log.Firewalled):
         if ircutils.isUserHostmask(msg.prefix) and not msg.command == 'NICK':
             self.nicksToHostmasks[msg.nick] = msg.prefix
         if 'batch' in msg.server_tags:
-            batch = msg.server_tags['batch']
-            assert batch in self.batches, \
-                    'Server references undeclared batch %s' % batch
-            self.batches[batch].messages.append(msg)
+            batch_name = msg.server_tags['batch']
+            assert batch_name in self.batches, \
+                'Server references undeclared batch %r' % batch_name
+            for batch in self.getParentBatches(msg):
+                batch.messages.append(msg)
         method = self.dispatchCommand(msg.command, msg.args)
         if method is not None:
             method(irc, msg)
@@ -768,6 +772,76 @@ class IrcState(IrcCommandDispatcher, log.Firewalled):
         """Returns the hostmask for a given nick."""
         return self.nicksToHostmasks[nick]
 
+    def getParentBatches(self, msg):
+        """Given an IrcMsg, returns a list of all batches that contain it,
+        innermost first.
+
+        Raises ValueError if ``msg`` is not in a batch;
+        or if it is in a batch that has already ended.
+        This restriction may be relaxed in the future.
+
+        This means that you should not call ``getParentBatches``
+        on a message that was already processed.
+
+        For example, assume Limnoria received the following::
+
+            :irc.host BATCH +outer example.com/foo
+            @batch=outer :irc.host BATCH +inner example.com/bar
+            @batch=inner :nick!user@host PRIVMSG #channel :Hi
+            @batch=outer :irc.host BATCH -inner
+            :irc.host BATCH -outer
+
+        If you call getParentBatches on any of the middle three messages,
+        you get ``[Batch(name='inner', ...), Batch(name='outer', ...)]``.
+        And if you call getParentBatches on either the first or the last
+        message, you get ``[Batch(name='outer', ...)]``
+
+        And you may only call `getParentBatches`` on the PRIVMSG
+        if only the first three messages were processed.
+        """
+        batch = msg.tagged('batch')
+        if not batch:
+            # msg is not a BATCH command
+            batch_name = msg.server_tags.get('batch')
+            if batch_name:
+                batch = self.batches.get(batch_name)
+                if not batch:
+                    raise ValueError(
+                        'Called getParentBatches for a message in a batch that '
+                        'already ended.'
+                    )
+            else:
+                raise ValueError(
+                    'Called getParentBatches for a message not in a batch.')
+
+        batches = []
+        while batch:
+            batches.append(batch)
+            batch = batch.parent_batch
+
+        return batches
+
+    def getClientTagDenied(self, tag):
+        """Returns whether the given tag is denied by the server, according
+        to its CLIENTTAGDENY policy.
+        This is only informative, and servers may still allow or deny tags
+        at their discretion.
+
+        For details, see the RPL_ISUPPORT section in
+        <https://ircv3.net/specs/extensions/message-tags>
+        """
+        tag = tag.lstrip("+")
+
+        denied_tags = self.supported.get('CLIENTTAGDENY')
+        if not denied_tags:
+            return False
+        denied_tags = denied_tags.split(',')
+        if '*' in denied_tags:
+            # All tags are denied by default, check the whitelist
+            return ('-' + tag) not in denied_tags
+        else:
+            return tag in denied_tags
+
     def do004(self, irc, msg):
         """Handles parsing the 004 reply
 
@@ -1017,8 +1091,20 @@ class IrcState(IrcCommandDispatcher, log.Firewalled):
         if msg.args[0].startswith('+'):
             batch_type = msg.args[1]
             batch_arguments = tuple(msg.args[2:])
-            self.batches[batch_name] = Batch(type=batch_type,
-                    arguments=batch_arguments, messages=[msg])
+
+            # Both are possibly None:
+            parent_batch_name = msg.server_tags.get("batch")
+            parent_batch = self.batches.get(parent_batch_name)
+
+            batch = Batch(
+                name=batch_name,
+                type=batch_type,
+                arguments=batch_arguments,
+                messages=[msg],
+                parent_batch=parent_batch
+            )
+            msg.tag('batch', batch)
+            self.batches[batch_name] = batch
         elif msg.args[0].startswith('-'):
             batch = self.batches.pop(batch_name)
             batch.messages.append(msg)
@@ -1241,7 +1327,8 @@ class Irc(IrcCommandDispatcher, log.Firewalled):
         else:
             msg_tags_str = ''
             msg_rest_str = msg_str
-        if len(msg_rest_str) > MAX_LINE_SIZE:
+        msg_rest_bytes = msg_rest_str.encode()
+        if len(msg_rest_bytes) > MAX_LINE_SIZE:
             # Yes, this violates the contract, but at this point it doesn't
             # matter.  That's why we gotta go munging in private attributes
             #
@@ -1250,7 +1337,19 @@ class Irc(IrcCommandDispatcher, log.Firewalled):
             # this issue, there's no fundamental reason to make it a
             # warning.
             log.debug('Truncating %r, message is too long.', msg)
-            msg._str = msg_tags_str + msg_rest_str[:MAX_LINE_SIZE-2] + '\r\n'
+
+            # Truncate to 512 bytes (minus 2 for '\r\n')
+            msg_rest_bytes = msg_rest_bytes[:MAX_LINE_SIZE-2]
+
+            # The above truncation may have truncated in the middle of a
+            # multi-byte character.
+            # I was about to write a UTF-8 decoder here just to trim them
+            # properly, but fortunately there is a neat trick to trim it
+            # while decoding: just ignore invalid bytes!
+            # https://stackoverflow.com/a/1820949/539465
+            msg_rest_str = msg_rest_bytes.decode(errors="ignore")
+
+            msg._str = msg_tags_str + msg_rest_str + '\r\n'
             msg._len = len(str(msg))
         # TODO: truncate tags
 
@@ -1523,6 +1622,10 @@ class Irc(IrcCommandDispatcher, log.Firewalled):
     To check if a capability was negotiated, use `irc.state.capabilities_ack`.
     """
 
+    REQUEST_EXPERIMENTAL_CAPABILITIES = set(['draft/multiline'])
+    """Like REQUEST_CAPABILITIES, but these capabilities are only requested
+    if supybot.protocols.irc.experimentalExtensions is enabled."""
+
     def _queueConnectMessages(self):
         if self.zombie:
             self.driver.die()
@@ -1875,9 +1978,12 @@ class Irc(IrcCommandDispatcher, log.Firewalled):
             # Normally at this point, self.state.capabilities_ack should be
             # empty; but let's just make sure we're not requesting the same
             # caps twice for no reason.
+            want_capabilities = self.REQUEST_CAPABILITIES
+            if conf.supybot.protocols.irc.experimentalExtensions():
+                want_capabilities |= self.REQUEST_EXPERIMENTAL_CAPABILITIES
             new_caps = (
                 set(self.state.capabilities_ls) &
-                self.REQUEST_CAPABILITIES -
+                want_capabilities -
                 self.state.capabilities_ack)
             # NOTE: Capabilities are requested in alphabetic order, because
             # sets are unordered, and their "order" is nondeterministic.
@@ -2039,22 +2145,42 @@ class Irc(IrcCommandDispatcher, log.Firewalled):
         self.afterConnect = True
         # Let's reset nicks in case we had to use a weird one.
         self.alternateNicks = conf.supybot.nick.alternates()[:]
+
+        self._setUmodes()
+
+    do377 = do422 = do376
+
+    def _setUmodes(self):
+        # Get the configured umodes
         umodes = conf.supybot.networks.get(self.network).umodes()
         if umodes == '':
             umodes = conf.supybot.protocols.irc.umodes()
+
+        # Add the bot mode if the server advertizes one;
+        # and if the configured umode doesn't already have it
+        # explicitly set or unset
+        bot_mode = self.state.supported.get("BOT")
+        if bot_mode and len(bot_mode) == 1:
+            if bot_mode not in umodes:
+                umodes += "+" + bot_mode
+
+        # Filter out umodes not supported by the server
         supported = self.state.supported.get('umodes')
         if supported:
             acceptedchars = supported.union('+-')
             umodes = ''.join([m for m in umodes if m in acceptedchars])
+
+        # Send the umodes
         if umodes:
             log.info('Sending user modes to %s: %s', self.network, umodes)
             self.sendMsg(ircmsgs.mode(self.nick, umodes))
-    do377 = do422 = do376
 
     def do43x(self, msg, problem):
         if not self.afterConnect:
+            self.triedNicks.add(self.nick)
             newNick = self._getNextNick()
-            assert newNick != self.nick
+            assert newNick != self.nick, \
+                (self.nick, self.alternateNicks, self.triedNicks)
             log.info('Got %s: %s %s.  Trying %s.',
                      msg.command, self.nick, problem, newNick)
             self.sendMsg(ircmsgs.nick(newNick))
