@@ -29,6 +29,7 @@
 
 import copy
 import pickle
+import textwrap
 import unittest.mock
 
 from supybot.test import *
@@ -419,6 +420,29 @@ class IrcStateTestCase(SupyTestCase):
         self.assertEqual(state.supported['chanmodes'],
                          frozenset('biklmnopstveI'))
 
+    def testClientTagDenied(self):
+        state = irclib.IrcState()
+
+        # By default, every tag is assumed allowed
+        self.assertEqual(state.getClientTagDenied('foo'), False)
+        self.assertEqual(state.getClientTagDenied('bar'), False)
+
+        # Blacklist
+        state.addMsg(self.irc, ircmsgs.IrcMsg(
+            ':example.org 005 mybot CLIENTTAGDENY=foo :are supported by this server'))
+        self.assertEqual(state.getClientTagDenied('foo'), True)
+        self.assertEqual(state.getClientTagDenied('bar'), False)
+        self.assertEqual(state.getClientTagDenied('+foo'), True)
+        self.assertEqual(state.getClientTagDenied('+bar'), False)
+
+        # Whitelist
+        state.addMsg(self.irc, ircmsgs.IrcMsg(
+            ':example.org 005 mybot CLIENTTAGDENY=*,-foo :are supported by this server'))
+        self.assertEqual(state.getClientTagDenied('foo'), False)
+        self.assertEqual(state.getClientTagDenied('bar'), True)
+        self.assertEqual(state.getClientTagDenied('+foo'), False)
+        self.assertEqual(state.getClientTagDenied('+bar'), True)
+
     def testShort004(self):
         state = irclib.IrcState()
         state.addMsg(self.irc, ircmsgs.IrcMsg(':coulomb.oftc.net 004 testnick coulomb.oftc.net hybrid-7.2.2+oftc1.6.8'))
@@ -869,6 +893,52 @@ class IrcTestCase(SupyTestCase):
         self.irc.feedMsg(ircmsgs.IrcMsg(command='372', args=['nick', 'some message']))
         self.irc.feedMsg(ircmsgs.IrcMsg(command='376', args=['nick']))
 
+    def testSetUmodes(self):
+        def assertSentModes(modes):
+            self.assertEqual(
+                self.irc.takeMsg(),
+                ircmsgs.IrcMsg(command='MODE', args=['test', modes]),
+            )
+
+        self.irc.reset()
+        while self.irc.takeMsg():
+            pass
+
+        self.irc.state.supported["BOT"] = "" # invalid
+        self.irc._setUmodes()
+        self.assertIsNone(self.irc.takeMsg())
+
+        self.irc.state.supported["BOT"] = "bB" # invalid too
+        self.irc._setUmodes()
+        self.assertIsNone(self.irc.takeMsg())
+
+        del self.irc.state.supported["BOT"]
+        self.irc._setUmodes()
+        self.assertIsNone(self.irc.takeMsg())
+
+        self.irc.state.supported["BOT"] = "B"
+        self.irc._setUmodes()
+        assertSentModes("+B")
+
+        self.irc.state.supported["BOT"] = "b"
+        self.irc._setUmodes()
+        assertSentModes("+b")
+
+        # merge with configured umodes
+        with conf.supybot.protocols.irc.umodes.context("+B"):
+            self.irc._setUmodes()
+            assertSentModes("+B+b")
+
+        # no duplicate if char is the same
+        with conf.supybot.protocols.irc.umodes.context("+b"):
+            self.irc._setUmodes()
+            assertSentModes("+b")
+
+        # no duplicate if explicitly disabled
+        with conf.supybot.protocols.irc.umodes.context("-b"):
+            self.irc._setUmodes()
+            assertSentModes("-b")
+
     def testMsgChannel(self):
         self.irc.reset()
 
@@ -966,7 +1036,100 @@ class IrcTestCase(SupyTestCase):
             self.irc.feedMsg(m4)
         finally:
             self.irc.removeCallback(c.name())
-        self.assertEqual(c.batch, irclib.Batch('netjoin', (), [m1, m2, m3, m4]))
+        self.assertEqual(
+            c.batch,
+            irclib.Batch('name', 'netjoin', (), [m1, m2, m3, m4], None),
+            repr(c.batch)
+        )
+
+    maxDiff = None
+
+    def testBatchNested(self):
+        self.irc.reset()
+        logs = textwrap.dedent('''
+            :irc.host BATCH +outer example.com/foo
+            @batch=outer :irc.host BATCH +inner example.com/bar
+            @batch=inner :nick!user@host PRIVMSG #channel :Hi
+            @batch=outer :irc.host BATCH -inner
+            :irc.host BATCH -outer
+        ''')
+        msgs = [ircmsgs.IrcMsg(s) for s in logs.split('\n') if s]
+
+        # Feed 'BATCH +outer', it should be added in state
+        self.irc.feedMsg(msgs[0])
+        outer = irclib.Batch('outer', 'example.com/foo', (), msgs[0:1], None)
+        self.assertEqual(self.irc.state.batches, {'outer': outer})
+        msg1 = self.irc.state.history[-1]
+        self.assertEqual(msg1.tagged('batch'), outer)
+        self.assertEqual(self.irc.state.getParentBatches(msg1), [outer])
+
+        # Feed 'BATCH +inner', it should be added in state
+        self.irc.feedMsg(msgs[1])
+        outer = irclib.Batch('outer', 'example.com/foo', (), msgs[0:2], None)
+        inner = irclib.Batch('inner', 'example.com/bar', (), msgs[1:2], outer)
+        self.assertIs(self.irc.state.batches['inner'].parent_batch,
+                      self.irc.state.batches['outer'])
+        self.assertEqual(dict(self.irc.state.batches),
+                         {'outer': outer, 'inner': inner})
+        msg2 = self.irc.state.history[-1]
+        self.assertEqual(msg2.tagged('batch'), inner)
+        self.assertEqual(self.irc.state.getParentBatches(msg2), [inner, outer])
+
+        # Feed 'PRIVMSG'
+        self.irc.feedMsg(msgs[2])
+        outer = irclib.Batch('outer', 'example.com/foo', (), msgs[0:3], None)
+        inner = irclib.Batch('inner', 'example.com/bar', (), msgs[1:3], outer)
+        self.assertIs(self.irc.state.batches['inner'].parent_batch,
+                      self.irc.state.batches['outer'])
+        self.assertEqual(self.irc.state.batches,
+                         {'outer': outer, 'inner': inner})
+        msg3 = self.irc.state.history[-1]
+        self.assertEqual(msg3.tagged('batch'), None)
+        self.assertEqual(self.irc.state.getParentBatches(msg3), [inner, outer])
+
+        # Feed 'BATCH -inner', it should be remove from state
+        self.irc.feedMsg(msgs[3])
+        outer = irclib.Batch('outer', 'example.com/foo', (), msgs[0:4], None)
+        inner = irclib.Batch('inner', 'example.com/bar', (), msgs[1:4], outer)
+        self.assertEqual(self.irc.state.batches, {'outer': outer})
+        self.assertIs(self.irc.state.history[-1].tagged('batch').parent_batch,
+                      self.irc.state.batches['outer'])
+        msg4 = self.irc.state.history[-1]
+        self.assertEqual(msg4.tagged('batch'), inner)
+        self.assertEqual(self.irc.state.getParentBatches(msg4), [inner, outer])
+
+        # Feed 'BATCH -outer', it should be remove from state
+        self.irc.feedMsg(msgs[4])
+        outer = irclib.Batch('outer', 'example.com/foo', (), msgs[0:5], None)
+        inner = irclib.Batch('inner', 'example.com/bar', (), msgs[1:4], outer)
+        self.assertEqual(self.irc.state.batches, {})
+        msg5 = self.irc.state.history[-1]
+        self.assertEqual(msg5.tagged('batch'), outer)
+        self.assertEqual(self.irc.state.getParentBatches(msg5), [outer])
+
+    def testTruncate(self):
+        self.irc = irclib.Irc('test')
+
+        while self.irc.takeMsg():
+            pass
+
+        # over 512 bytes
+        msg = ircmsgs.IrcMsg(command='PRIVMSG', args=('#test2', 'é'*256))
+        self.irc.sendMsg(msg)
+        m = self.irc.takeMsg()
+        remaining_payload = 'é' * ((512 - len('PRIVMSG #test2 :\r\n'))//2)
+        self.assertEqual(
+                str(m), 'PRIVMSG #test2 :%s\r\n' % remaining_payload)
+
+        # over 512 bytes, make sure it doesn't truncate in the middle of
+        # a character
+        msg = ircmsgs.IrcMsg(command='PRIVMSG', args=('#test', 'é'*256))
+        self.irc.sendMsg(msg)
+        m = self.irc.takeMsg()
+        remaining_payload = 'é' * ((512 - len('PRIVMSG #test :\r\n'))//2)
+        self.assertEqual(
+                str(m), 'PRIVMSG #test :%s\r\n' % remaining_payload)
+
 
 class SaslTestCase(SupyTestCase, CapNegMixin):
     def setUp(self):
